import logging
import os
import pathlib
from dataclasses import replace
from itertools import chain
from subprocess import CompletedProcess, TimeoutExpired
from typing import Any, NamedTuple, cast
from unittest import mock

import pytest
from pytest_mock import MockerFixture, MockType

from competitive_verifier import oj
from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import (
    JudgeStatus,
    Problem,
    VerifcationTimeoutError,
)
from competitive_verifier.models import (
    TestCaseFile as SystemTestCaseFile,
)
from competitive_verifier.models import (
    TestCaseProvider as SystemTestCaseProvider,
)
from competitive_verifier.oj.oj_test import (
    OjExecInfo,
    OjTestArguments,
    OjTestcaseResult,
    OjTestResult,
    gnu_time_message,
    measure_command,
    single_case,
    special_judge,
    summarize,
)
from competitive_verifier.oj.problem import (
    AOJProblem,
    LibraryCheckerProblem,
    YukicoderProblem,
)
from tests import LogComparer

OJ_TEST_MODULE = "competitive_verifier.oj.oj_test"


def make_result(
    name: str,
    status: JudgeStatus,
    elapsed: float = 1.25,
    memory: float | None = None,
    exitcode: int | None = 0,
) -> OjTestcaseResult:
    return OjTestcaseResult(
        name=name,
        elapsed=elapsed,
        exitcode=exitcode,
        memory=memory,
        status=status,
        input=pathlib.Path(),
        expected=pathlib.Path(),
        answer="",
    )


@pytest.fixture
def mock_measure(
    mocker: MockerFixture,
    request: pytest.FixtureRequest,
):
    assert isinstance(request.param, OjExecInfo)

    mocker.patch(
        "competitive_verifier.oj.oj_test.measure_command",
        return_value=request.param,
    )
    return request.param


@pytest.fixture
def mock_generate_test_cases(mocker: MockerFixture):
    mocker.patch.object(
        LibraryCheckerProblem,
        "generate_test_cases",
    )


@pytest.fixture
def mock_judge(
    mocker: MockerFixture,
    request: pytest.FixtureRequest,
) -> Problem:
    judge = getattr(request, "param", None)

    if judge is not None:
        assert isinstance(judge, bool)
        mocker.patch(
            "competitive_verifier.oj.oj_test.special_judge",
            return_value=judge,
        )
        mocker.patch.object(
            LibraryCheckerProblem,
            "checker",
            return_value=pathlib.Path("/anywhere/mockcheck"),
            new_callable=mocker.PropertyMock,
        )
        return LibraryCheckerProblem(problem_id="aplusb")
    return AOJProblem(problem_id="1")


class SingleCaseParams(NamedTuple):
    name: str
    inbytes: bytes
    mock_measure: OjExecInfo
    expected: OjTestcaseResult
    expected_log: list[LogComparer]
    outbytes: bytes
    error: float | None = None
    mle: float | None = None
    mock_judge: bool | None = None


test_single_case_params: list[SingleCaseParams] = [
    SingleCaseParams(
        name="default",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="default",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("default: start"),
            LogComparer("default: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="Nones",
        inbytes=b"",
        outbytes=b"",
        mock_measure=OjExecInfo(answer=None, elapsed=1.25, memory=None, returncode=0),
        expected=make_result(
            name="Nones",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("Nones: start"),
            LogComparer("Nones: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="memory",
        inbytes=b"",
        outbytes=b"19\n20\n",
        mock_measure=OjExecInfo(
            answer="19\n20\n", elapsed=1.25, memory=10.7, returncode=0
        ),
        expected=make_result(
            name="memory",
            memory=10.7,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("memory: start"),
            LogComparer(
                "memory: \x1b[32mAC\x1b[39m, time: 1.250000 sec, memory: 10.700000 MB"
            ),
        ],
    ),
    SingleCaseParams(
        name="spaces",
        inbytes=b"foo \t bar\nbaz\t \r\nhoge\r\nfuga\n",
        outbytes=b"1\n",
        mock_measure=OjExecInfo(answer="2\n", elapsed=1.25, memory=None, returncode=0),
        expected=make_result(
            name="spaces",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("spaces: start"),
            LogComparer(
                "spaces:input: "
                "\x1b[1mfoo\x1b[0m\x1b[2m_\\t_\x1b[0m\x1b[1mbar\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mbaz\x1b[0m\x1b[2m\\t_\x1b[0m\x1b[2m(trailing whitespace)\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mhoge\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mfuga\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer("spaces:answer: \x1b[1m2\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("spaces:expected: \x1b[1m1\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("spaces: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="empty-lines",
        inbytes=b"\n\n\n\n",
        outbytes=b"1\n",
        mock_measure=OjExecInfo(answer="2\n", elapsed=1.25, memory=None, returncode=0),
        expected=make_result(
            name="empty-lines",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("empty-lines: start"),
            LogComparer("empty-lines:input: \x1b[2m\\n\\n\\n\\n\x1b[0m"),
            LogComparer("empty-lines:answer: \x1b[1m2\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("empty-lines:expected: \x1b[1m1\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("empty-lines: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="judge-AC",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_judge=True,
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="judge-AC",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("judge-AC: start"),
            LogComparer("judge-AC: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="judge-WA",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_judge=False,
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="judge-WA",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("judge-WA: start"),
            LogComparer("judge-WA:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("judge-WA:answer: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("judge-WA:expected: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("judge-WA: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="judge-WA-no-expected-out",
        inbytes=b"abc\n",
        outbytes=b"",
        mock_judge=False,
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="judge-WA-no-expected-out",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("judge-WA-no-expected-out: start"),
            LogComparer(
                "judge-WA-no-expected-out:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            LogComparer(
                "judge-WA-no-expected-out:answer: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            LogComparer("judge-WA-no-expected-out:expected: \x1b[2m(empty)\x1b[0m"),
            LogComparer(
                "judge-WA-no-expected-out: \x1b[31mWA\x1b[39m, time: 1.250000 sec"
            ),
        ],
    ),
    SingleCaseParams(
        name="check_output-WA",
        inbytes=b"abc\n",
        outbytes=b"ABC",
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="check_output-WA",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("check_output-WA: start"),
            LogComparer(
                "This was AC if spaces and newlines were ignored.",
                level=logging.WARNING,
            ),
            LogComparer(
                "check_output-WA:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer("check_output-WA:answer: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer(
                "check_output-WA:expected: \x1b[1mABC\x1b[0m\x1b[2m(no trailing newline)\x1b[0m",
            ),
            LogComparer("check_output-WA: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="check_output-AC",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="check_output-AC",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("check_output-AC: start"),
            LogComparer("check_output-AC: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-AC-equals",
        inbytes=b"Pi\n",
        outbytes=b"3.14159265358\n",
        error=1e-11,
        mock_measure=OjExecInfo(
            answer="3.14159265358\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-AC-equals",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-AC-equals: start"),
            LogComparer("error-AC-equals: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-AC-long",
        inbytes=b"Pi\n",
        outbytes=b"3.141592653589793238462643383279502884197169399375105820974944592307816406286\n",
        error=1e-11,
        mock_measure=OjExecInfo(
            answer="3.14159265358\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-AC-long",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-AC-long: start"),
            LogComparer("error-AC-long: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-AC-short",
        inbytes=b"Pi\n",
        outbytes=b"3.1415926536\n",
        error=1e-11,
        mock_measure=OjExecInfo(
            answer="3.14159265358\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-AC-short",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-AC-short: start"),
            LogComparer("error-AC-short: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-small-AC-abs",
        inbytes=b"Planck mass\n",
        outbytes=b"0.0000000217640\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer="0.0000000217647\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-small-AC-abs",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-small-AC-abs: start"),
            LogComparer("error-small-AC-abs: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-small-AC-abs-short",
        inbytes=b"Planck mass\n",
        outbytes=b"0.000000021764\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer="0.0000000217647\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-small-AC-abs-short",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-small-AC-abs-short: start"),
            LogComparer(
                "error-small-AC-abs-short: \x1b[32mAC\x1b[39m, time: 1.250000 sec"
            ),
        ],
    ),
    SingleCaseParams(
        name="error-small-WA-abs",
        inbytes=b"Planck mass\n",
        outbytes=b"0.0000000217640\n",
        error=1e-13,
        mock_measure=OjExecInfo(
            answer="0.0000000217647\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-small-WA-abs",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("error-small-WA-abs: start"),
            LogComparer(
                "error-small-WA-abs:input: "
                "\x1b[1mPlanck\x1b[0m\x1b[2m_\x1b[0m\x1b[1mmass\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "error-small-WA-abs:answer: \x1b[1m0.0000000217647\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "error-small-WA-abs:expected: \x1b[1m0.0000000217640\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer("error-small-WA-abs: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-large-AC-rel",
        inbytes=b"Planck temp\n",
        outbytes=b"141678400000000000000000000000000\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer="141678400000000000000000000000000\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="error-large-AC-rel",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-large-AC-rel: start"),
            LogComparer("error-large-AC-rel: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-large-AC-rel-diff",
        inbytes=b"Planck temp\n",
        outbytes=b"141678400000000000000000000000000\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer="141678400000021987654321987654321\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="error-large-AC-rel-diff",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-large-AC-rel-diff: start"),
            LogComparer(
                "error-large-AC-rel-diff: \x1b[32mAC\x1b[39m, time: 1.250000 sec"
            ),
        ],
    ),
    SingleCaseParams(
        name="error-large-WA-rel",
        inbytes=b"Planck temp\n",
        outbytes=b"141678400000000000000000000000000\n",
        error=1e-13,
        mock_measure=OjExecInfo(
            answer="141678400000021987654321987654321\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="error-large-WA-rel",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("error-large-WA-rel: start"),
            LogComparer(
                "error-large-WA-rel:input: "
                "\x1b[1mPlanck\x1b[0m\x1b[2m_\x1b[0m\x1b[1mtemp\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "error-large-WA-rel:answer: \x1b[1m141678400000021987654321987654321\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "error-large-WA-rel:expected: \x1b[1m141678400000000000000000000000000\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer("error-large-WA-rel: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-same",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\nGH\n",
        mock_measure=OjExecInfo(
            answer="ABC\nDEF\nGH\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-same",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("lines-same: start"),
            LogComparer("lines-same: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-crlf",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\nGH\n",
        mock_measure=OjExecInfo(
            answer="ABC\r\nDEF\r\nGH\r\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-crlf",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("lines-crlf: start"),
            LogComparer("lines-crlf: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-diff",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\nGH\n",
        mock_measure=OjExecInfo(
            answer="ABC\nDDF\nGH\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-diff",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("lines-diff: start"),
            LogComparer("lines-diff:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer(
                "lines-diff:answer: "
                "\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mDDF\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mGH\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "lines-diff:expected: "
                "\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mGH\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer("lines-diff: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-error-diff-len",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer="ABC\nDEF\n3.14159265\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-error-diff-len",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("lines-error-diff-len: start"),
            LogComparer("lines-error-diff-len: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-error-grid",
        inbytes=b"abc\n",
        outbytes=b"ABC CD\nDEF 3.141592653589\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer="ABC CD\nDEF 3.14159265\n3.14159265\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="lines-error-grid",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("lines-error-grid: start"),
            LogComparer("lines-error-grid: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-grid-diff",
        inbytes=b"abc\n",
        outbytes=b"ABC CD\nDEF 3.141592653589\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer="ABC CD\nDEF 3.14159265 3.14\n3.14159265\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="lines-grid-diff",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("lines-grid-diff: start"),
            LogComparer("lines-grid-diff:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer(
                "lines-grid-diff:answer: "
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "lines-grid-diff:expected: "
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.141592653589\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265358979\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer("lines-grid-diff: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-grid-diff-lines",
        inbytes=b"abc\n",
        outbytes=b"ABC CD\nDEF 3.141592653589\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer="ABC CD\nDEF 3.14159265 3.14\n3.14159265\nno\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="lines-grid-diff-lines",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("lines-grid-diff-lines: start"),
            LogComparer(
                "lines-grid-diff-lines:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            LogComparer(
                "lines-grid-diff-lines:answer: "
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mno\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "lines-grid-diff-lines:expected: "
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.141592653589\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265358979\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "lines-grid-diff-lines: \x1b[31mWA\x1b[39m, time: 1.250000 sec"
            ),
        ],
    ),
    SingleCaseParams(
        name="error-str-same",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        error=1e-3,
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-str-same",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("error-str-same: start"),
            LogComparer("error-str-same: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-str-WA",
        inbytes=b"abc\n",
        outbytes=b"DEF\n",
        error=1e-3,
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-str-WA",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("error-str-WA: start"),
            LogComparer("error-str-WA:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("error-str-WA:answer: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("error-str-WA:expected: \x1b[1mDEF\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("error-str-WA: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="TLE",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=None
        ),
        expected=make_result(
            name="TLE",
            exitcode=None,
            status=JudgeStatus.TLE,
        ),
        expected_log=[
            LogComparer("TLE: start"),
            LogComparer("TLE:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("TLE:expected: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("TLE: \x1b[31mTLE\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="RE",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=1
        ),
        expected=make_result(
            name="RE",
            exitcode=1,
            status=JudgeStatus.RE,
        ),
        expected_log=[
            LogComparer("RE: start"),
            LogComparer("RE:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("RE:expected: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("RE: \x1b[31mRE\x1b[39m, time: 1.250000 sec, return code: 1"),
        ],
    ),
    SingleCaseParams(
        name="MLE-NotMeasure",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mle=128,
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="MLE-NotMeasure",
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("MLE-NotMeasure: start"),
            LogComparer("MLE-NotMeasure: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="MLE-Safe",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mle=128,
        mock_measure=OjExecInfo(answer="ABC\n", elapsed=1.25, memory=128, returncode=0),
        expected=make_result(
            name="MLE-Safe",
            memory=128,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            LogComparer("MLE-Safe: start"),
            LogComparer(
                "MLE-Safe: \x1b[32mAC\x1b[39m, time: 1.250000 sec, memory: 128.000000 MB"
            ),
        ],
    ),
    SingleCaseParams(
        name="MLE",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mle=128,
        mock_measure=OjExecInfo(
            answer="ABC\n", elapsed=1.25, memory=128.1, returncode=1
        ),
        expected=make_result(
            name="MLE",
            exitcode=1,
            memory=128.1,
            status=JudgeStatus.MLE,
        ),
        expected_log=[
            LogComparer("MLE: start"),
            LogComparer("MLE:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("MLE:answer: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("MLE:expected: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer(
                "MLE: \x1b[31mMLE\x1b[39m,"
                " time: 1.250000 sec, memory: 128.100000 MB, return code: 1"
            ),
        ],
    ),
    SingleCaseParams(
        name="long-lines",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer="""ea226088-3cc7-4d33-985c-00004dfcdccb
6f791930-4b08-4618-a110-32be65332f08
87f28765-03bb-4aa3-ad2c-5e781aa60ca0
2c006877-aeae-4b24-88f1-29351ac40ad7
73380cec-9775-4b5d-9f86-b0bba61f2608
""",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="long-lines",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("long-lines: start"),
            LogComparer("long-lines:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer(
                "long-lines:answer: "
                "\x1b[1mea226088-3cc7-4d33-9\x1b[0m\x1b[2m... (145 chars) "
                "...\x1b[0m\x1b[1md-9f86-b0bba61f2608\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer("long-lines:expected: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("long-lines: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="long-text",
        inbytes=b"abc\r\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer="ea226088-3cc7-4d33-985c-00004dfcdccb6f791930-4b08-4618-a110-32be65332f08"
            "87f28765-03bb-4aa3-ad2c-5e781aa60ca09b063b0f-7b48-4a1d-a49f-4e8524ace9174"
            "2e6d61b-9656-4cb6-be7a-1a5d80224390961bed54-a32d-47c7-bb82-ae4591ed8ccd\r\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="long-text",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("long-text: start"),
            LogComparer(
                "long-text:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "long-text:answer: "
                "\x1b[1mea226088-3cc7-4d33-9\x1b[0m\x1b[2m... (178 chars) ...\x1b[0m"
                "\x1b[1m-bb82-ae4591ed8ccd\x1b[0m\x1b[2m\\r\\n\x1b[0m",
            ),
            LogComparer("long-text:expected: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer("long-text: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="long-lines-break",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer="""ea226088-3cc7-4d33-\n985c-00004dfcdccb
6f791930-4b08-4618-a110-32be65332f08
87f28765-03bb-4aa3-ad2c-5e781aa60ca0
2c006877-aeae-4b24-88f1-29351ac40ad7
73380cec-9775-4b5d\n-9f86-b0bba61f2608
""",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="long-lines-break",
            status=JudgeStatus.WA,
        ),
        expected_log=[
            LogComparer("long-lines-break: start"),
            LogComparer("long-lines-break:input: \x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            LogComparer(
                "long-lines-break:answer: "
                "\x1b[1mea226088-3cc7-4d33-\x1b[0m\x1b[2m\\n\x1b[0m\x1b[2m... (147 chars) "
                "...\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1m-9f86-b0bba61f2608\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            LogComparer(
                "long-lines-break:expected: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            LogComparer("long-lines-break: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
]


@pytest.mark.usefixtures("mock_generate_test_cases")
@pytest.mark.parametrize(
    SingleCaseParams._fields,
    test_single_case_params,
    ids=[f"{i}:{t.name}" for i, t in enumerate(test_single_case_params)],
    indirect=["mock_measure", "mock_judge"],
)
def test_single_case(
    name: str,
    inbytes: bytes,
    outbytes: bytes,
    error: float | None,
    mle: float | None,
    expected: OjTestcaseResult,
    expected_log: list[LogComparer],
    caplog: pytest.LogCaptureFixture,
    mock_judge: Problem,
    mock_measure: OjExecInfo,
    testtemp: pathlib.Path,
):
    caplog.set_level(logging.NOTSET)
    input_path = testtemp / "infile"
    input_path.write_bytes(inbytes)

    output_path = testtemp / "outfile"
    output_path.write_bytes(outbytes)

    result = single_case(
        name,
        test_input_path=input_path,
        test_output_path=output_path,
        args=OjTestArguments(
            command="ls ~",
            problem=mock_judge,
            error=error,
            mle=mle,
            tle=None,
        ),
    )

    expected.answer = mock_measure.answer or ""
    expected.input = input_path
    expected.expected = output_path
    assert result == expected
    assert caplog.records == expected_log


def test_single_case_error(
    mocker: MockerFixture,
    mock_judge: Problem,
    caplog: pytest.LogCaptureFixture,
    testtemp: pathlib.Path,
    subtests: pytest.Subtests,
):
    caplog.set_level(logging.NOTSET)
    input_path = testtemp / "infile"
    input_path.write_bytes(b"ABC")

    output_path = testtemp / "outfile"
    output_path.write_bytes(b"abc")

    expected = OjTestcaseResult(
        name="error",
        input=input_path,
        answer="",
        expected=output_path,
        status=JudgeStatus.RE,
        elapsed=0,
        exitcode=255,
    )

    mocker.patch("competitive_verifier.oj.gnu.time_command", return_value="time")
    for cmd in ["", ":no_exists:"]:
        with subtests.test(msg=cmd if cmd else "<empty>"):
            caplog.clear()
            assert (
                single_case(
                    "error",
                    test_input_path=input_path,
                    test_output_path=output_path,
                    args=OjTestArguments(
                        command=cmd,
                        problem=mock_judge,
                        error=None,
                        mle=None,
                        tle=None,
                    ),
                )
                == expected
            )
            assert caplog.records == [
                LogComparer("error: start"),
                LogComparer(
                    "Failed to run: OjTestArguments(command='" + cmd + "', "
                    "problem=AOJProblem.from_url('http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1'), "
                    "tle=None, mle=None, error=None, env=None, deadline=inf)",
                    level=logging.ERROR,
                    github=GitHubMessageParams(),
                ),
            ]
        with subtests.test(msg="Exception"):
            caplog.clear()
            mocker.patch("subprocess.Popen", side_effect=Exception)
            assert (
                single_case(
                    "error",
                    test_input_path=input_path,
                    test_output_path=output_path,
                    args=OjTestArguments(
                        command="git",
                        problem=mock_judge,
                        error=None,
                        mle=None,
                        tle=None,
                    ),
                )
                == expected
            )
            assert caplog.records[0] == LogComparer("error: start")
            assert caplog.records[2] == LogComparer(
                "Failed to run: OjTestArguments(command='git', "
                "problem=AOJProblem.from_url('http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1'), "
                "tle=None, mle=None, error=None, env=None, deadline=inf)",
                level=logging.ERROR,
                github=GitHubMessageParams(),
            )

            assert caplog.records[1].msg.endswith("is not executable.")


class GnuTimeMessageParam(NamedTuple):
    has_gnu_time: bool
    system: str
    mle: float | None
    expected: list[LogComparer]


test_gnu_time_message_params: list[GnuTimeMessageParam] = [
    GnuTimeMessageParam(True, "Darwin", 5.5, []),
    GnuTimeMessageParam(True, "Darwin", None, []),
    GnuTimeMessageParam(True, "Windows", 2, []),
    GnuTimeMessageParam(True, "Windows", None, []),
    GnuTimeMessageParam(
        False,
        "Darwin",
        5.5,
        [
            LogComparer(
                "[HINT]: You can install GNU time with: $ brew install gnu-time",
                logging.INFO,
                github=GitHubMessageParams(),
            ),
            LogComparer(
                "--mle is used but GNU time does not exist",
                logging.WARNING,
                github=GitHubMessageParams(),
            ),
        ],
    ),
    GnuTimeMessageParam(
        False,
        "Darwin",
        None,
        [
            LogComparer(
                "[HINT]: You can install GNU time with: $ brew install gnu-time",
                logging.INFO,
                github=GitHubMessageParams(),
            ),
        ],
    ),
    GnuTimeMessageParam(
        False,
        "Windows",
        2,
        [
            LogComparer(
                "--mle is used but GNU time does not exist",
                logging.WARNING,
                github=GitHubMessageParams(),
            ),
        ],
    ),
    GnuTimeMessageParam(
        False,
        "Windows",
        None,
        [],
    ),
]


@pytest.mark.parametrize(GnuTimeMessageParam._fields, test_gnu_time_message_params)
def test_gnu_time_message(
    has_gnu_time: bool,
    system: str,
    mle: float | None,
    expected: list[LogComparer],
    mock_judge: Problem,
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
):
    caplog.set_level(0)
    mocker.patch(
        "competitive_verifier.oj.gnu.time_command",
        return_value="time" if has_gnu_time else None,
    )
    mocker.patch("platform.system", return_value=system)

    gnu_time_message(
        OjTestArguments("dummy", problem=mock_judge, tle=None, mle=mle, error=None)
    )
    assert caplog.records == expected


def test_timeout(
    mocker: MockerFixture,
    mock_judge: Problem,
    mock_perf_counter: MockType,
):
    single_case_mock = mocker.patch(
        "competitive_verifier.oj.oj_test.single_case",
        return_value=make_result(
            name="default",
            elapsed=1.25,
            memory=None,
            status=JudgeStatus.AC,
        ),
    )
    mocker.patch.object(
        mock_judge,
        "iter_system_cases",
        return_value=[
            SystemTestCaseFile(
                name=str(i),
                input_path=pathlib.Path(str(i)),
                output_path=pathlib.Path(str(i)),
            )
            for i in range(100)
        ],
    )

    with pytest.raises(VerifcationTimeoutError):
        oj.test(
            problem=mock_judge,
            command="dummy",
            env=None,
            tle=None,
            mle=None,
            error=None,
            deadline=2.99,
        )

    assert single_case_mock.call_count == 3


class TestMeasureCommand:
    @pytest.fixture
    def mock_run(self, mocker: MockerFixture, request: pytest.FixtureRequest):
        mocker.patch("shutil.which", side_effect=lambda cmd: cmd == "dummy_command")  # pyright: ignore[reportUnknownLambdaType]
        ret = getattr(request, "param", None)
        if ret is None:
            ret = CompletedProcess[str]("dummy_command 1", returncode=0)

        if isinstance(ret, CompletedProcess):
            return mocker.patch("subprocess.run", return_value=ret)
        if isinstance(ret, Exception):
            return mocker.patch("subprocess.run", side_effect=ret)
        raise NotImplementedError

    @pytest.mark.usefixtures("mock_perf_counter")
    @pytest.mark.parametrize(
        ("mock_run", "expected"),
        [
            (
                CompletedProcess[str]("dummy_command 1", returncode=0, stdout="ABC\n"),
                OjExecInfo(answer="ABC\n", elapsed=1.0, memory=None, returncode=0),
            ),
            (
                CompletedProcess[str]("dummy_command 1", returncode=1, stdout="ABC\n"),
                OjExecInfo(answer="ABC\n", elapsed=1.0, memory=None, returncode=1),
            ),
            (
                TimeoutExpired("dummy_command 1", timeout=2, output="ABC\n"),
                OjExecInfo(answer=None, elapsed=1.0, memory=None, returncode=None),
            ),
        ],
        indirect=["mock_run"],
    )
    def test_result(
        self,
        expected: OjExecInfo,
        mock_run: MockType,
        testtemp: pathlib.Path,
        subtests: pytest.Subtests,
    ):
        with subtests.test(msg="gnu_time=False"):
            assert measure_command("dummy_command", gnu_time=False) == expected
            assert mock_run.call_args.args == (["dummy_command"],)
            assert mock_run.call_args.kwargs["start_new_session"] is False
            mock_run.reset_mock()

        with (
            subtests.test(msg="gnu_time=True but no gnu_time"),
            mock.patch("competitive_verifier.oj.gnu.time_command", return_value=None),
        ):
            assert measure_command("dummy_command", gnu_time=True) == expected
            assert mock_run.call_args.args == (["dummy_command"],)
            assert mock_run.call_args.kwargs["start_new_session"] is False
            mock_run.reset_mock()

        with (
            subtests.test(msg="gnu_time=True"),
            mock.patch(
                "competitive_verifier.oj.gnu.time_command", return_value="dummy_time"
            ),
            mock.patch(
                "competitive_verifier.oj.gnu._GnuTimeRunnerImpl.get_memory",
                return_value=22.5,
            ),
        ):
            assert measure_command("dummy_command", gnu_time=True) == replace(
                expected, memory=22.5
            )
            assert mock_run.call_args.args == (
                [
                    "dummy_time",
                    "-f",
                    "%M",
                    "-o",
                    str(testtemp / "1" / "gnu_time_report.txt"),
                    "--",
                    "dummy_command",
                ],
            )
            assert mock_run.call_args.kwargs["start_new_session"] is True
            mock_run.reset_mock()

    @pytest.mark.parametrize(
        ("cmd", "expected"),
        [
            ("dummy_command -- -a", ["dummy_command", "--", "-a"]),
            (["dummy_command", "--", "-a"], ["dummy_command", "--", "-a"]),
        ],
    )
    def test_command(
        self,
        cmd: str | list[str],
        expected: list[str],
        mock_run: MockType,
    ):
        measure_command(cmd, gnu_time=False)
        assert mock_run.call_args.args == (expected,)

    @pytest.mark.parametrize(
        ("env", "expected_env"),
        [
            (
                None,
                None,
            ),
            (
                {},
                {"FOO": "1", "BAR": "2"},
            ),
            (
                {"BAZ": "3"},
                {"FOO": "1", "BAR": "2", "BAZ": "3"},
            ),
            (
                {"BAZ": "3", "BAR": "4"},
                {"FOO": "1", "BAR": "4", "BAZ": "3"},
            ),
        ],
    )
    def test_env(
        self,
        env: dict[str, str] | None,
        expected_env: dict[str, str],
        mocker: MockerFixture,
        mock_run: MockType,
    ):
        mocker.patch.dict(os.environ, {"FOO": "1", "BAR": "2"}, clear=True)
        measure_command("dummy_command -- a", env=env, gnu_time=False)

        mock_run.assert_called_once()
        assert mock_run.call_args.kwargs["env"] == expected_env


test_oj_test_params: dict[str, tuple[dict[str, Any], OjTestArguments]] = {
    "default": (
        {
            "problem": LibraryCheckerProblem(problem_id="example"),
            "command": "ls .",
            "tle": 2,
            "error": None,
            "mle": 128,
            "env": None,
        },
        OjTestArguments(
            command="ls .",
            tle=2,
            mle=128,
            error=None,
            problem=LibraryCheckerProblem(problem_id="example"),
        ),
    ),
    "with_env": (
        {
            "problem": YukicoderProblem(problem_no=10),
            "command": ["ls", "."],
            "tle": 30,
            "error": None,
            "mle": 256,
            "env": {"TOKEN": "Dummy"},
        },
        OjTestArguments(
            command=["ls", "."],
            tle=30,
            mle=256,
            error=None,
            problem=YukicoderProblem(problem_no=10),
            env={"TOKEN": "Dummy"},
        ),
    ),
}


@pytest.mark.parametrize(
    ("args", "expected"),
    test_oj_test_params.values(),
    ids=test_oj_test_params.keys(),
)
def test_oj_test(
    mocker: MockerFixture,
    args: dict[str, Any],
    expected: OjTestArguments,
):
    mocker.patch(
        "competitive_verifier.oj.problem.LibraryCheckerProblem.checker_exe_name",
        "mockcheck",
    )
    run = mocker.patch("competitive_verifier.oj.oj_test._run")

    oj.test(**args)

    run.assert_called_once_with(expected)


class MockCasesProblem(SystemTestCaseProvider):
    cases: list[SystemTestCaseFile]

    def __init__(self, cases: list[SystemTestCaseFile]) -> None:
        self.cases = cases

    def download_system_cases(self) -> Any:
        raise NotImplementedError

    def iter_system_cases(self) -> list[SystemTestCaseFile]:
        return self.cases


def test_oj_test_run(mocker: MockerFixture, subtests: pytest.Subtests):
    case_count = 0

    def infinite_case(*args: Any, **kwargs: Any):
        nonlocal case_count
        while True:
            case_count += 1
            return make_result(name=f"case{case_count}", status=JudgeStatus.AC)

    mocker.patch(
        "competitive_verifier.oj.oj_test.single_case",
        side_effect=infinite_case,
    )
    mock_summarize = mocker.patch("competitive_verifier.oj.oj_test.summarize")

    with subtests.test(msg="empty"):
        oj.test(
            problem=MockCasesProblem([]),
            command="dummy",
            env=None,
            tle=None,
            mle=None,
            error=None,
        )
        mock_summarize.assert_called_once_with([])

    with subtests.test(msg="three"):
        mock_summarize.reset_mock()
        oj.test(
            problem=MockCasesProblem(
                [
                    SystemTestCaseFile(
                        name=f"c{i}",
                        input_path=pathlib.Path(),
                        output_path=pathlib.Path(),
                    )
                    for i in range(3)
                ]
            ),
            command="dummy",
            env=None,
            tle=None,
            mle=None,
            error=None,
        )
        mock_summarize.assert_called_once_with(
            [
                make_result(name="case1", status=JudgeStatus.AC),
                make_result(name="case2", status=JudgeStatus.AC),
                make_result(name="case3", status=JudgeStatus.AC),
            ]
        )


def test_compare_answer_too_large_error(
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
):
    mocker.patch("competitive_verifier.oj.oj_test.summarize")

    oj.test(
        problem=MockCasesProblem([]),
        command="dummy",
        env=None,
        tle=None,
        mle=None,
        error=0.9999,
    )
    assert not caplog.records

    oj.test(
        problem=MockCasesProblem([]),
        command="dummy",
        env=None,
        tle=None,
        mle=None,
        error=1.0001,
    )

    assert caplog.records == [
        LogComparer(
            "the tolerance is too large: relative = 1.0001",
            logging.WARNING,
            github=GitHubMessageParams(),
        ),
    ]


EXPECTED_CASES = cast("list[OjTestcaseResult]", ...)
test_summarize_params = [
    pytest.param(
        [],
        OjTestResult(
            is_success=True,
            elapsed=0,
            slowest=-1.0,
            heaviest=-1.0,
            testcases=EXPECTED_CASES,
        ),
        [
            LogComparer("\x1b[32mSUCCESS\x1b[39m 0 cases"),
        ],
        id="empty",
    ),
    pytest.param(
        [
            make_result(
                f"case{i}",
                status=JudgeStatus.AC,
                elapsed=i * 0.5,
                memory=i * 1.5,
            )
            for i in [9, 6, 5, 8, 1, 2, 3, 4, 7, 0]
        ],
        OjTestResult(
            is_success=True,
            elapsed=45 / 2,
            slowest=9 / 2,
            heaviest=9 * 1.5,
            testcases=EXPECTED_CASES,
        ),
        [
            LogComparer("slowest: 4.500000 sec  (for case9)"),
            LogComparer("max memory: 13.500000 MB  (for case9)"),
            LogComparer("\x1b[32mSUCCESS\x1b[39m 10 cases"),
        ],
        id="shuffle",
    ),
    pytest.param(
        [
            make_result(
                f"case{i}",
                status=JudgeStatus.AC,
                elapsed=i * 0.5,
            )
            for i in [9, 6, 5, 8, 1, 2, 3, 4, 7, 0]
        ],
        OjTestResult(
            is_success=True,
            elapsed=45 / 2,
            slowest=9 / 2,
            heaviest=-1.0,
            testcases=EXPECTED_CASES,
        ),
        [
            LogComparer("slowest: 4.500000 sec  (for case9)"),
            LogComparer("\x1b[32mSUCCESS\x1b[39m 10 cases"),
        ],
        id="no_memory",
    ),
    pytest.param(
        [
            make_result(
                f"case{i}",
                status=st,
                elapsed=i * 0.5,
                memory=i * 1.5,
            )
            for i, st in enumerate(
                chain(
                    *(
                        [st] * cnt
                        for st, cnt in zip(
                            list(JudgeStatus), range(1, 10000), strict=False
                        )
                    )
                )
            )
        ],
        OjTestResult(
            is_success=False,
            elapsed=(3 * len(JudgeStatus) - 1) * (3 * len(JudgeStatus)) / 2 / 2,
            slowest=(3 * len(JudgeStatus) - 1) / 2,
            heaviest=(3 * len(JudgeStatus) - 1) * 1.5,
            testcases=EXPECTED_CASES,
        ),
        [
            LogComparer("slowest: 7.000000 sec  (for case14)"),
            LogComparer("max memory: 21.000000 MB  (for case14)"),
            LogComparer(
                "\x1b[31mFAILURE\x1b[39m 1 AC, 2 WA, 3 RE, 4 TLE, 5 MLE / 15 cases",
            ),
        ],
        id="statuses",
    ),
]


@pytest.mark.parametrize(("history", "expected", "expected_log"), test_summarize_params)
def test_summarize(
    history: list[OjTestcaseResult],
    expected: OjTestResult,
    expected_log: list[tuple[str, int, str]],
    caplog: pytest.LogCaptureFixture,
):
    caplog.set_level(0)
    expected.testcases = history
    assert summarize(history) == expected
    assert caplog.records == expected_log


test_special_judge_params = [
    (
        None,
        None,
        False,
        [
            LogComparer(
                "judge's output: \x1b[2m(empty)\x1b[0m",
                10,
            )
        ],
    ),
    (
        "a b c",
        1,
        False,
        [
            LogComparer(
                "judge's output: "
                "\x1b[1ma\x1b[0m\x1b[2m_\x1b[0m\x1b[1mb\x1b[0m\x1b[2m_\x1b[0m\x1b[1mc\x1b[0m"
                "\x1b[2m(no trailing newline)\x1b[0m",
                10,
            )
        ],
    ),
    (
        "ABC\n",
        0,
        True,
        [
            LogComparer(
                "judge's output: \x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m",
                10,
            )
        ],
    ),
]


@pytest.mark.parametrize(
    ("answer", "returncode", "expected", "expected_log"),
    test_special_judge_params,
)
def test_special_judge(
    answer: str | None,
    returncode: int | None,
    expected: bool,
    expected_log: list[LogComparer],
    mocker: MockerFixture,
    testtemp: pathlib.Path,
    caplog: pytest.LogCaptureFixture,
):
    caplog.set_level(0)
    mock_measure = mocker.patch(
        "competitive_verifier.oj.oj_test.measure_command",
        return_value=OjExecInfo(
            answer=answer, elapsed=0, memory=None, returncode=returncode
        ),
    )
    assert (
        special_judge(
            "/special/check",
            "a b c\n",
            input_path=testtemp / "input",
            expected_output_path=testtemp / "expected",
        )
        == expected
    )

    check_cmd = [
        "/special/check",
        str(testtemp / "input"),
        str(testtemp / "1/actual.out"),
        str(testtemp / "expected"),
    ]
    mock_measure.assert_called_once_with(check_cmd)
    assert caplog.records[0] == LogComparer(
        f"$ {check_cmd!s}",
        10,
    )
    assert caplog.records[1:] == expected_log



import pathlib
from argparse import ArgumentParser
from logging import getLogger
from typing import Literal

from pydantic import Field

from competitive_verifier import config
from competitive_verifier.arg import (
    IgnoreErrorArguments,
    IncludeExcludeArguments,
    ResultJsonArguments,
    VerboseArguments,
    VerifyFilesJsonArguments,
    WriteSummaryArguments,
)
from competitive_verifier.inout import MergeResult
from competitive_verifier.models import (
    VerificationInput,
)

from .builder import DocumentBuilder

logger = getLogger(__name__)


def get_default_docs_dir() -> pathlib.Path:
    default_docs_dir = config.get_config_dir() / "docs"
    oj_verify_docs_dir = pathlib.Path(".verify-helper/docs")
    if not default_docs_dir.exists() and oj_verify_docs_dir.exists():
        return oj_verify_docs_dir
    return default_docs_dir


class Docs(
    IncludeExcludeArguments,
    WriteSummaryArguments,
    IgnoreErrorArguments,
    ResultJsonArguments,
    VerifyFilesJsonArguments,
    VerboseArguments,
):
    subcommand: Literal["docs"] = Field(
        default="docs",
        description="Create documents",
    )
    docs: pathlib.Path | None = None
    destination: pathlib.Path

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        destination = config.get_config_dir() / "_jekyll"
        parser.add_argument(
            "--docs",
            type=pathlib.Path,
            help=f"Document settings directory. default: {get_default_docs_dir().as_posix()}",
        )
        parser.add_argument(
            "--destination",
            type=pathlib.Path,
            default=destination,
            help=f"Output directory for markdown document. default: {destination.as_posix()}",
        )

    def _run(self) -> bool:
        logger.debug("arguments:%s", self)
        logger.info("path of verify_files_json=%s", self.verify_files_json)
        logger.info("path of result_json=%s", [str(p) for p in self.result_json])
        verifications = VerificationInput.parse_file_relative(self.verify_files_json)

        result = MergeResult(
            subcommand="merge-result",
            result_json=self.result_json,
        ).merge()

        if self.write_summary:
            self.write_result(result)

        logger.debug("verifications=%s", verifications)
        logger.debug("result=%s", result)

        return DocumentBuilder(
            verifications=verifications,
            result=result,
            docs_dir=self.docs or get_default_docs_dir(),
            destination_dir=self.destination,
            include=self.include,
            exclude=self.exclude,
        ).build()



import os
import pathlib

COMPETITIVE_VERIFY_CONFIG_PATH = "COMPETITIVE_VERIFY_CONFIG_PATH"


def get_config_dir():
    return pathlib.Path(
        os.getenv(COMPETITIVE_VERIFY_CONFIG_PATH, ".competitive-verifier")
    )


def get_cache_dir():
    return get_config_dir() / "cache"


def get_problem_cache_dir() -> pathlib.Path:
    return get_cache_dir() / "problems"



import logging
import os
import pathlib
from datetime import datetime, timedelta

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import summary
from competitive_verifier.arg import WriteSummaryArguments
from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import (
    FileResult,
    JudgeStatus,
    ResultStatus,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.models import TestcaseResult as CaseResult
from tests import LogComparer

test_to_human_str_seconds_params: list[tuple[timedelta, str]] = [
    (timedelta(hours=2, minutes=8, seconds=6, milliseconds=13), "2h 8m"),
    (timedelta(minutes=8, seconds=6, milliseconds=13), "8m 6s"),
    (timedelta(seconds=16, milliseconds=13), "16s"),
    (timedelta(seconds=6, milliseconds=3), "6.0s"),
    (timedelta(seconds=6, milliseconds=13), "6.0s"),
    (timedelta(seconds=6, milliseconds=213), "6.2s"),
    (timedelta(milliseconds=1), "1ms"),
    (timedelta(milliseconds=12), "12ms"),
    (timedelta(milliseconds=123), "123ms"),
]


@pytest.mark.parametrize(
    ("td", "expected"),
    test_to_human_str_seconds_params,
    ids=str,
)
def test_to_human_str_seconds(td: timedelta, expected: str):
    assert summary.to_human_str_seconds(td.total_seconds()) == expected


test_to_human_str_mega_bytes_params: list[tuple[float, str]] = [
    (0.000123456789, "0MB"),
    (0.00123456, "0.00123MB"),
    (0.01234567, "0.0123MB"),
    (0.12345678, "0.123MB"),
    (1.23456789, "1.23MB"),
    (12.3456789, "12.3MB"),
    (123.456789, "123MB"),
    (1234.56789, "1234MB"),
    (12345.6789, "12345MB"),
]


@pytest.mark.parametrize(
    ("megabytes", "expected"),
    test_to_human_str_mega_bytes_params,
)
def test_to_human_str_mega_bytes(megabytes: float, expected: str):
    assert summary.to_human_str_mega_bytes(megabytes) == expected


class MockWriteSummaryArguments(WriteSummaryArguments):
    def run(self) -> bool: ...


def test_no_summary(mocker: MockerFixture):
    mocker.patch.dict(os.environ, {}, clear=True)
    mock_summary = mocker.patch.object(summary, "write_summary")
    MockWriteSummaryArguments(write_summary=False).write_result(
        VerifyCommandResult(files={}, total_seconds=1)
    )
    mock_summary.assert_not_called()


def test_summary_not_exist(mocker: MockerFixture, caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.NOTSET)
    mocker.patch.dict(os.environ, {}, clear=True)
    mock_summary = mocker.patch.object(summary, "write_summary")

    MockWriteSummaryArguments(write_summary=True).write_result(
        VerifyCommandResult(files={}, total_seconds=1)
    )
    mock_summary.assert_not_called()
    assert caplog.records == [
        LogComparer(
            "write_summary=True but not found $GITHUB_STEP_SUMMARY",
            logging.WARNING,
            github=GitHubMessageParams(),
        )
    ]


test_summary_params = [
    (
        VerifyCommandResult(
            total_seconds=4.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            verification_name="ver",
                            heaviest=31.432342,
                            slowest=0.420321,
                            elapsed=1,
                            status=ResultStatus.FAILURE,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                            testcases=[
                                CaseResult(
                                    name="case01",
                                    status=JudgeStatus.TLE,
                                    elapsed=0.2,
                                    memory=30,
                                ),
                                CaseResult(
                                    name="case02",
                                    status=JudgeStatus.WA,
                                    elapsed=0.4,
                                    memory=32,
                                ),
                            ],
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.FAILURE,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            verification_name="fi",
                            heaviest=31.432342,
                            slowest=0.420321,
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                            testcases=[
                                CaseResult(
                                    name="case01",
                                    status=JudgeStatus.AC,
                                    elapsed=0.2,
                                    memory=30,
                                ),
                                CaseResult(
                                    name="case02",
                                    status=JudgeStatus.AC,
                                    elapsed=0.4,
                                    memory=32,
                                ),
                            ],
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("hoge.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            heaviest=31.432342,
                            slowest=0.420321,
                            elapsed=1,
                            status=ResultStatus.FAILURE,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                            testcases=[
                                CaseResult(
                                    name="case01",
                                    status=JudgeStatus.RE,
                                    elapsed=0.2,
                                    memory=30,
                                ),
                                CaseResult(
                                    name="case02",
                                    status=JudgeStatus.MLE,
                                    elapsed=0.4,
                                    memory=32,
                                ),
                            ],
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("piyo.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("old.py"): FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ],
                ),
            },
        ),
        r"""# ❌ Verification result

- ✔&nbsp;&nbsp;All test case results are `success`
- ❌&nbsp;&nbsp;Test case results containts `failure`
- ⚠&nbsp;&nbsp;Test case results containts `skipped`


## Results
|📝&nbsp;&nbsp;File|✔<br>Passed|❌<br>Failed|⚠<br>Skipped|∑<br>Total|⏳<br>Elapsed|🦥<br>Slowest|🐘<br>Heaviest|
|:---|:---:|:---:|:---:|:---:|:---:|:---:|:---:|
|_**Sum**_|3|3|1|7|4.2s|-|-|
|||||||||
|❌&nbsp;&nbsp;foo/bar.py|-|2|-|2|3.0s|420ms|31.4MB|
|✔&nbsp;&nbsp;foo/baz.py|2|-|-|2|3.0s|420ms|31.4MB|
|❌&nbsp;&nbsp;hoge.py|1|1|-|2|3.0s|420ms|31.4MB|
|⚠&nbsp;&nbsp;piyo.py|-|-|1|1|2.0s|-|-|
## Past results
|📝&nbsp;&nbsp;File|✔<br>Passed|❌<br>Failed|⚠<br>Skipped|∑<br>Total|⏳<br>Elapsed|🦥<br>Slowest|🐘<br>Heaviest|
|:---|:---:|:---:|:---:|:---:|:---:|:---:|:---:|
|✔&nbsp;&nbsp;old.py|2|-|-|2|4.0s|-|-|
## Failed tests

### foo/bar.py

|env|name|status|Elapsed|Memory|
|:---|:---|:---:|:---:|:---:|
|ver|case01|TLE|200ms|30MB|
|ver|case02|WA|400ms|32MB|
### hoge.py

|env|name|status|Elapsed|Memory|
|:---|:---|:---:|:---:|:---:|
||case01|RE|200ms|30MB|
||case02|MLE|400ms|32MB|
""",
    ),
    (
        VerifyCommandResult(
            total_seconds=11.12,
            files={
                pathlib.Path("hoge.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                    ]
                ),
                pathlib.Path("piyo.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
            },
        ),
        r"""# ⚠ Verification result

- ✔&nbsp;&nbsp;All test case results are `success`
- ❌&nbsp;&nbsp;Test case results containts `failure`
- ⚠&nbsp;&nbsp;Test case results containts `skipped`


## Results
|📝&nbsp;&nbsp;File|✔<br>Passed|❌<br>Failed|⚠<br>Skipped|∑<br>Total|⏳<br>Elapsed|🦥<br>Slowest|🐘<br>Heaviest|
|:---|:---:|:---:|:---:|:---:|:---:|:---:|:---:|
|_**Sum**_|1|-|1|2|11s|-|-|
|||||||||
|✔&nbsp;&nbsp;hoge.py|1|-|-|1|2.0s|-|-|
|⚠&nbsp;&nbsp;piyo.py|-|-|1|1|2.0s|-|-|
""",
    ),
    (
        VerifyCommandResult(
            total_seconds=11.12,
            files={
                pathlib.Path("hoge.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                    ]
                ),
                pathlib.Path("piyo.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ],
                ),
            },
        ),
        r"""# ✔ Verification result

- ✔&nbsp;&nbsp;All test case results are `success`
- ❌&nbsp;&nbsp;Test case results containts `failure`
- ⚠&nbsp;&nbsp;Test case results containts `skipped`


## Results
|📝&nbsp;&nbsp;File|✔<br>Passed|❌<br>Failed|⚠<br>Skipped|∑<br>Total|⏳<br>Elapsed|🦥<br>Slowest|🐘<br>Heaviest|
|:---|:---:|:---:|:---:|:---:|:---:|:---:|:---:|
|_**Sum**_|2|-|-|2|11s|-|-|
|||||||||
|✔&nbsp;&nbsp;hoge.py|1|-|-|1|2.0s|-|-|
|✔&nbsp;&nbsp;piyo.py|1|-|-|1|2.0s|-|-|
""",
    ),
    (
        VerifyCommandResult(
            total_seconds=11.12,
            files={
                pathlib.Path("hoge.py"): FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                    ],
                ),
                pathlib.Path("piyo.py"): FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ],
                ),
            },
        ),
        r"""# ✔ Verification result

- ✔&nbsp;&nbsp;All test case results are `success`
- ❌&nbsp;&nbsp;Test case results containts `failure`
- ⚠&nbsp;&nbsp;Test case results containts `skipped`


## Past results
|📝&nbsp;&nbsp;File|✔<br>Passed|❌<br>Failed|⚠<br>Skipped|∑<br>Total|⏳<br>Elapsed|🦥<br>Slowest|🐘<br>Heaviest|
|:---|:---:|:---:|:---:|:---:|:---:|:---:|:---:|
|✔&nbsp;&nbsp;hoge.py|1|-|-|1|2.0s|-|-|
|✔&nbsp;&nbsp;piyo.py|1|-|-|1|2.0s|-|-|
""",
    ),
]


@pytest.mark.parametrize(
    ("verify_command_result", "expected"),
    test_summary_params,
    ids=lambda s: f"len(expected)={len(s)}" if isinstance(s, str) else s,
)
def test_summary(
    verify_command_result: VerifyCommandResult,
    expected: str,
    mocker: MockerFixture,
    testtemp: pathlib.Path,
):
    tmp = testtemp / "summary.md"
    mocker.patch.dict(
        os.environ,
        {"GITHUB_STEP_SUMMARY": str(tmp)},
        clear=True,
    )
    MockWriteSummaryArguments(write_summary=True).write_result(verify_command_result)

    assert tmp.read_text(encoding="utf-8") == expected.replace("\r\n", "\n")



import importlib.resources
import pathlib
from typing import cast

_DOC_USAGE_PATH = "doc_usage.txt"


_ROOT = importlib.resources.files(cast("str", __package__))


def doc_usage(
    *,
    markdown_dir_path: pathlib.Path,
    repo_name: str,
) -> str:
    template = _ROOT / _DOC_USAGE_PATH
    return (
        template.read_text(encoding="utf-8")
        .replace("{{{{{markdown_dir_path}}}}}", markdown_dir_path.as_posix())
        .replace("{{{{{repository}}}}}", repo_name)
    )


def jekyll_files() -> dict[str, bytes]:
    return {
        path: (_ROOT / "jekyll" / path).read_bytes()
        for path in [
            "_layouts/page.html",
            "_layouts/document.html",
            "_layouts/multidoc.html",
            "_layouts/toppage.html",
            "_includes/head-custom.html",
            "_includes/head-custom2.html",
            "_includes/mathjax/mathjax.html",
            "_includes/mathjax/mathjax2.html",
            "_includes/mathjax/mathjax3.html",
            "_includes/code.html",
            "_includes/code_and_testcases.html",
            "_includes/highlight_additional.html",
            "_includes/highlight/highlight_header.html",
            "_includes/dependencies.html",
            "_includes/document_header.html",
            "_includes/document_body.html",
            "_includes/document_footer.html",
            "_includes/multidoc_body.html",
            "_includes/multidoc_header.html",
            "_includes/toppage_header.html",
            "_includes/toppage_body.html",
            "_includes/toppage_footer.html",
            "assets/css/default.scss",
            "assets/css/code.scss",
            "assets/js/code.js",
            "Gemfile",
        ]
    }



from enum import Enum


class ResultStatus(str, Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    SKIPPED = "skipped"

    @property
    def status(self) -> "ResultStatus":
        return self


class JudgeStatus(Enum):
    AC = "AC"
    WA = "WA"
    RE = "RE"
    TLE = "TLE"
    MLE = "MLE"



import datetime
import json
import pathlib
import random
import shutil

import pytest

from competitive_verifier import app
from competitive_verifier.models import (
    FileResult,
    JudgeStatus,
    ResultStatus,
    VerificationResult,
)
from competitive_verifier.models import TestcaseResult as _TestcaseResult
from competitive_verifier.verify import verifier

from .data.integration_data import IntegrationData


@pytest.mark.integration
@pytest.mark.usefixtures("additional_path")
@pytest.mark.order(-500)
class TestCommandVerfy:
    @pytest.mark.usefixtures("mock_verification")
    def test_mock_dump(self):
        for _ in range(20):
            command_result = verifier.VerifyCommandResult(
                total_seconds=3,
                files={
                    pathlib.Path("foo/bar.py"): FileResult(
                        verifications=[
                            VerificationResult(
                                verification_name="name1",
                                elapsed=random.randint(10, 242),
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.datetime.fromtimestamp(
                                    random.randint(1000000000, 3000000000)
                                ),
                            ),
                            VerificationResult(
                                verification_name="name2",
                                elapsed=random.randint(10, 242),
                                status=ResultStatus.SKIPPED,
                                heaviest=random.randint(1000, 3000),
                                slowest=random.randint(234, 1234),
                                last_execution_time=datetime.datetime.fromtimestamp(
                                    random.randint(1000000000, 3000000000)
                                ),
                                testcases=[
                                    _TestcaseResult(
                                        name="case01",
                                        status=JudgeStatus.AC,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                    _TestcaseResult(
                                        name="case02",
                                        status=JudgeStatus.RE,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                    _TestcaseResult(
                                        name="case03",
                                        status=JudgeStatus.WA,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                    _TestcaseResult(
                                        name="case04",
                                        status=JudgeStatus.TLE,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                ],
                            ),
                        ]
                    ),
                    pathlib.Path("foo/baz2.py"): FileResult(
                        verifications=[
                            VerificationResult(
                                verification_name="name1",
                                elapsed=random.randint(10, 242),
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.datetime.fromtimestamp(
                                    random.randint(1000000000, 3000000000)
                                ),
                            ),
                        ]
                    ),
                },
            )
            assert json.loads(command_result.model_dump_json(exclude_none=True)) == {
                "total_seconds": 2547.12,
                "files": {
                    "foo/bar.py": {
                        "verifications": [
                            {
                                "verification_name": "name1",
                                "status": "success",
                                "elapsed": 9948.0,
                                "last_execution_time": "2004-10-09T00:26:13.890000+02:00",
                            },
                            {
                                "verification_name": "name2",
                                "status": "skipped",
                                "elapsed": 2238.0,
                                "slowest": 223.0,
                                "heaviest": 753.0,
                                "testcases": [
                                    {
                                        "name": "case01",
                                        "status": "AC",
                                        "elapsed": 1.33,
                                        "memory": 32.38,
                                    },
                                    {
                                        "name": "case02",
                                        "status": "RE",
                                        "elapsed": 4.73,
                                        "memory": 9.88,
                                    },
                                    {
                                        "name": "case03",
                                        "status": "WA",
                                        "elapsed": 2.4,
                                        "memory": 35.43,
                                    },
                                    {
                                        "name": "case04",
                                        "status": "TLE",
                                        "elapsed": 2.42,
                                        "memory": 22.41,
                                    },
                                ],
                                "last_execution_time": "2026-11-12T22:34:03.760000-11:00",
                            },
                        ],
                        "newest": True,
                    },
                    "foo/baz2.py": {
                        "verifications": [
                            {
                                "verification_name": "name1",
                                "status": "success",
                                "elapsed": 1239.0,
                                "last_execution_time": "1975-10-05T09:39:58.780000-09:00",
                            }
                        ],
                        "newest": True,
                    },
                },
            }

    @pytest.mark.usefixtures("mock_verification")
    def test_verify(
        self,
        integration_data: IntegrationData,
    ):
        verify = integration_data.config_dir_path / "verify.json"
        result = integration_data.config_dir_path / "result.json"
        shutil.rmtree(integration_data.config_dir_path / "cache", ignore_errors=True)

        parsed = app.ArgumentParser().parse(
            ["verify", "--verify-json", str(verify), "--output", str(result)]
        )
        assert isinstance(parsed, app.Verify)
        assert parsed.run()

        assert (
            json.loads(pathlib.Path(result).read_bytes())
            == integration_data.expected_verify_result()
        )



import glob
import json
import os
import pathlib
import posixpath
import re
import subprocess
import sys
import urllib.parse
import zipfile
from abc import abstractmethod
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from io import BytesIO
from logging import getLogger
from typing import ClassVar, Optional, TypeVar

import requests

from competitive_verifier import config
from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import (
    Problem,
    TestCaseData,
    TestCaseFile,
    TestCaseProvider,
)

logger = getLogger(__name__)


class NotLoggedInError(RuntimeError):
    pass


class _BaseProblem(Problem):
    def iter_system_cases(self) -> Iterator[TestCaseFile]:
        return iter_testcases(directory=self.test_directory)

    def download_system_cases(self) -> Iterable[TestCaseData] | bool:
        test_directory = self.test_directory

        if test_directory.exists() and any(test_directory.iterdir()):
            logger.info("download:already exists: %s", self.url)
            return True

        self.problem_directory.mkdir(parents=True, exist_ok=True)

        samples = list(self._download_cases())

        # Check samples
        if not samples:
            logger.error(
                "Sample not found",
                extra={"github": GitHubMessageParams()},
            )
            return False

        # write samples to files
        save_testcases(samples, directory=test_directory)
        return samples

    @abstractmethod
    def _download_cases(self) -> Iterable[TestCaseData]: ...


class LibraryCheckerProblem(Problem):
    checker_exe_name: ClassVar[str] = (
        "checker.exe" if sys.platform == "win32" else "checker"
    )

    def __init__(self, *, problem_id: str):
        self.problem_id = problem_id
        self._source_directory = None

    def __hash__(self) -> int:
        return hash((self.problem_id, self.repo_path))

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, LibraryCheckerProblem):
            return False
        return self.problem_id == value.problem_id and self.repo_path == value.repo_path

    @property
    def repo_path(self):
        return config.get_cache_dir() / "library-checker-problems"

    def iter_system_cases(self) -> Iterator[TestCaseFile]:
        inputs: dict[str, pathlib.Path] = {}
        outputs: dict[str, pathlib.Path] = {}
        for path in self.source_directory.glob("in/*.in"):
            inputs[path.stem] = path
        for path in self.source_directory.glob("out/*.out"):
            outputs[path.stem] = path
        return merge_testcase_files(inputs, outputs)

    def download_system_cases(self) -> bool:
        self.problem_directory.mkdir(parents=True, exist_ok=True)
        self.generate_test_cases()
        return True

    @property
    def checker(self) -> pathlib.Path | None:
        return self.source_directory / self.checker_exe_name

    def generate_test_cases(self) -> None:
        self.update_cloned_repository()
        path = self.repo_path

        spec = str(self.source_directory / "info.toml")
        command = [sys.executable, str(path / "generate.py"), spec]
        logger.info("$ %s", " ".join(command))
        try:
            subprocess.check_call(command, stdout=sys.stderr, stderr=sys.stderr)
        except subprocess.CalledProcessError:
            logger.exception(
                "the generate.py failed: check https://github.com/yosupo06/library-checker-problems/issues",
                extra={"github": GitHubMessageParams()},
            )
            raise

    @property
    def source_directory(self):
        if self._source_directory is None:
            problem_id = self.problem_id
            info_tomls = list(
                self.repo_path.glob(f"**/{glob.escape(problem_id)}/info.toml")
            )
            if len(info_tomls) != 1:
                raise RuntimeError(f"the problem {problem_id!r} not found or broken")
            self._source_directory = info_tomls[0].parent
        return self._source_directory

    @property
    def url(self) -> str:
        return f"https://judge.yosupo.jp/problem/{self.problem_id}"

    @classmethod
    def from_url(cls, url: str) -> Optional["LibraryCheckerProblem"]:
        # example: https://judge.yosupo.jp/problem/unionfind
        result = urllib.parse.urlparse(url)
        if result.scheme in ("", "http", "https") and result.netloc in (
            "judge.yosupo.jp",
            "old.yosupo.jp",
        ):
            m = re.match(r"/problem/(\w+)/?", result.path)
            if m:
                return cls(problem_id=m.group(1))
        return None

    _is_repository_updated = False

    def update_cloned_repository(self) -> None:
        if self._is_repository_updated:
            return

        try:
            subprocess.check_call(
                ["git", "--version"],  # noqa: S607
                stdout=sys.stderr,
                stderr=sys.stderr,
            )
        except FileNotFoundError:
            logger.exception(
                "git command not found",
                exc_info=False,
                extra={"github": GitHubMessageParams()},
            )
            raise

        path = self.repo_path
        if not path.exists():
            # init the problem repository
            url = "https://github.com/yosupo06/library-checker-problems"
            logger.info("$ git clone %s %s", url, path)
            subprocess.check_call(
                ["git", "clone", url, str(path)],  # noqa: S607
                stdout=sys.stderr,
                stderr=sys.stderr,
            )
        else:
            # sync the problem repository
            logger.info("$ git -C %s pull", path)
            subprocess.check_call(
                ["git", "-C", str(path), "pull"],  # noqa: S607
                stdout=sys.stderr,
                stderr=sys.stderr,
            )

        self._is_repository_updated = True


class _YukicoderProblemNo(int):
    def __new__(cls, value: int):
        return super().__new__(cls, value)

    def __str__(self) -> str:
        return "no/" + super().__str__()


class _YukicoderProblemId(int):
    def __new__(cls, value: int):
        return super().__new__(cls, value)


class YukicoderProblem(_BaseProblem):
    problem: _YukicoderProblemNo | _YukicoderProblemId

    def __init__(self, *, problem_no: int | None = None, problem_id: int | None = None):
        if problem_no is not None:
            self.problem = _YukicoderProblemNo(problem_no)
        elif problem_id is not None:
            self.problem = _YukicoderProblemId(problem_id)
        else:
            raise ValueError("Needs problem_no or problem_id")

    def _download_cases(self) -> list[TestCaseData]:
        """Download yukicoder problem.

        Raises:
            NotLoggedInError: If the `cargo metadata` command fails
        """
        headers: dict[str, str] | None = None
        if yukicoder_token := os.environ.get("YUKICODER_TOKEN"):
            headers = {"Authorization": f"Bearer {yukicoder_token}"}

        if not self._is_logged_in(headers=headers):
            raise NotLoggedInError("Required: $YUKICODER_TOKEN environment variable")
        url = f"{self.url}/testcase.zip"
        resp = requests.get(url, headers=headers, allow_redirects=True, timeout=10)

        with zipfile.ZipFile(BytesIO(resp.content)) as fh:
            inputs: dict[str, bytes] = {}
            outputs: dict[str, bytes] = {}
            for filename in fh.namelist():
                if filename.endswith("/"):
                    continue
                file = fh.read(filename)
                path = pathlib.Path(filename)
                if filename.startswith("test_in/"):
                    inputs[path.stem] = file
                elif filename.startswith("test_out/"):
                    outputs[path.stem] = file
            return [
                TestCaseData(name=name, input_data=i, output_data=o)
                for name, i, o in enumerate_inouts(inputs, outputs)
            ]

    @property
    def url(self) -> str:
        return f"https://yukicoder.me/problems/{self.problem}"

    @classmethod
    def from_url(cls, url: str) -> Optional["YukicoderProblem"]:
        # example: https://yukicoder.me/problems/no/499
        # example: http://yukicoder.me/problems/1476
        result = urllib.parse.urlparse(url)
        dirname, basename = posixpath.split(_normpath(result.path))
        if result.scheme in ("", "http", "https") and result.netloc == "yukicoder.me":
            try:
                n = int(basename)
            except ValueError:
                pass
            else:
                if dirname == "/problems/no":
                    return cls(problem_no=n)
                if dirname == "/problems":
                    return cls(problem_id=n)
        return None

    def _is_logged_in(self, *, headers: dict[str, str] | None = None) -> bool:
        url = "https://yukicoder.me"
        resp = requests.get(url, headers=headers, allow_redirects=True, timeout=10)
        resp.raise_for_status()
        return "login-btn" not in str(resp.content)


class AOJProblem(_BaseProblem):
    def __init__(self, *, problem_id: str):
        self.problem_id = problem_id

    def _download_cases(self) -> Iterable[TestCaseData]:
        return AOJProblem.download_cases(self.problem_id)

    @staticmethod
    def download_cases(problem_id: str) -> Iterable[TestCaseData]:
        # get header
        # reference: http://developers.u-aizu.ac.jp/api?key=judgedat%2Ftestcases%2F%7BproblemId%7D%2Fheader_GET
        url = f"https://judgedat.u-aizu.ac.jp/testcases/{problem_id}/header"
        resp = requests.get(url, allow_redirects=True, timeout=10)
        resp.raise_for_status()
        header_res = json.loads(resp.text)

        # get testcases via the official API
        for header in header_res["headers"]:
            # NOTE: the endpoints are not same to http://developers.u-aizu.ac.jp/api?key=judgedat%2Ftestcases%2F%7BproblemId%7D%2F%7Bserial%7D_GET since the json API often says "..... (terminated because of the limitation)"
            # NOTE: even when using https://judgedat.u-aizu.ac.jp/testcases/PROBLEM_ID/SERIAL, there is the 1G limit (see https://twitter.com/beet_aizu/status/1194947611100188672)
            serial = header["serial"]
            url = f"https://judgedat.u-aizu.ac.jp/testcases/{problem_id}/{serial}"

            resp_in = requests.get(url + "/in", allow_redirects=True, timeout=10)
            resp_in.raise_for_status()
            resp_out = requests.get(url + "/out", allow_redirects=True, timeout=10)
            resp_out.raise_for_status()

            yield TestCaseData(
                header["name"],
                resp_in.content,
                resp_out.content,
            )

    @property
    def url(self) -> str:
        return f"http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id={self.problem_id}"

    @classmethod
    def from_url(cls, url: str) -> Optional["AOJProblem"]:
        result = urllib.parse.urlparse(url)

        # example: http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1169
        # example: http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=DSL_1_A&lang=jp
        querystring = urllib.parse.parse_qs(result.query)
        if (
            result.scheme in ("", "http", "https")
            and result.netloc == "judge.u-aizu.ac.jp"
            and _normpath(result.path) == "/onlinejudge/description.jsp"
            and querystring.get("id")
            and len(querystring["id"]) == 1
        ):
            (n,) = querystring["id"]
            return cls(problem_id=n)

        # example: https://onlinejudge.u-aizu.ac.jp/challenges/sources/JAG/Prelim/2881
        # example: https://onlinejudge.u-aizu.ac.jp/courses/library/4/CGL/3/CGL_3_B
        m = re.match(
            r"^/(challenges|courses)/(sources|library/\d+|lesson/\d+)/(\w+)/(\w+)/(\w+)$",
            _normpath(result.path),
        )
        if (
            result.scheme in ("", "http", "https")
            and result.netloc == "onlinejudge.u-aizu.ac.jp"
            and m
        ):
            n = m.group(5)
            return cls(problem_id=n)

        # example: https://onlinejudge.u-aizu.ac.jp/problems/0423
        # example: https://onlinejudge.u-aizu.ac.jp/problems/CGL_3_B
        m = re.match(r"^/problems/(\w+)$", _normpath(result.path))
        if (
            result.scheme in ("", "http", "https")
            and result.netloc == "onlinejudge.u-aizu.ac.jp"
            and m
        ):
            n = m.group(1)
            return cls(problem_id=n)

        return None


class AOJArenaProblem(_BaseProblem):
    def __init__(self, *, arena_id: str, alphabet: str):
        if len(alphabet) != 1 or not alphabet.isupper():
            raise ValueError(arena_id, alphabet)
        self.arena_id = arena_id
        self.alphabet = alphabet

        self._problem_id: str | None = None

    def get_problem_id(self) -> str:
        if self._problem_id is None:
            url = f"https://judgeapi.u-aizu.ac.jp/arenas/{self.arena_id}/problems"
            resp = requests.get(url, allow_redirects=True, timeout=10)
            resp.raise_for_status()
            problems = json.loads(resp.text)
            for problem in problems:
                if problem["id"] == self.alphabet:
                    p = problem["problemId"]
                    logger.debug("problem: %s", p)
                    self._problem_id = p
                    return p
            raise ValueError("Problem is not found.")
        return self._problem_id

    def _download_cases(self) -> Iterable[TestCaseData]:
        return AOJProblem.download_cases(self.get_problem_id())

    @property
    def url(self) -> str:
        return f"https://onlinejudge.u-aizu.ac.jp/services/room.html#{self.arena_id}/problems/{self.alphabet}"

    @classmethod
    def from_url(cls, url: str) -> Optional["AOJArenaProblem"]:
        # example: https://onlinejudge.u-aizu.ac.jp/services/room.html#RitsCamp19Day2/problems/A
        result = urllib.parse.urlparse(url)
        if (
            result.scheme in ("", "http", "https")
            and result.netloc == "onlinejudge.u-aizu.ac.jp"
            and _normpath(result.path) == "/services/room.html"
        ):
            fragment = result.fragment.split("/")
            if len(fragment) == 3 and fragment[1] == "problems":  # noqa: PLR2004
                return cls(arena_id=fragment[0], alphabet=fragment[2].upper())
        return None


@dataclass
class LocalProblem(TestCaseProvider):
    path: pathlib.Path

    def download_system_cases(self) -> Iterable[TestCaseData] | bool:
        return bool(any(self.iter_system_cases()))

    def iter_system_cases(self) -> Iterable[TestCaseFile]:
        return iter_testcases(directory=self.path, recursive=True)


def _normpath(path: str) -> str:
    """A wrapper of posixpath.normpath.

    posixpath.normpath doesn't collapse a leading duplicated slashes.
    """
    path = posixpath.normpath(path)
    if path.startswith("//"):
        path = "/" + path.lstrip("/")
    return path


def _subclasses_recursive(cls: type[Problem]) -> Iterable[type[Problem]]:
    yield from (children := cls.__subclasses__())
    for ch in children:
        yield from _subclasses_recursive(ch)


def problem_from_url(url: str) -> Problem | None:
    for ch in set(_subclasses_recursive(Problem)):
        if (problem := ch.from_url(url)) is not None:
            return problem
    return None


_InOut = TypeVar("_InOut")


def enumerate_inouts(
    inputs: dict[str, _InOut],
    outputs: dict[str, _InOut],
) -> Iterator[tuple[str, _InOut, _InOut]]:
    common_keys = inputs.keys() & outputs.keys()
    if len(inputs) != len(common_keys) or len(outputs) != len(common_keys):
        logger.warning("dangling output case")

    if len(common_keys) == 0:
        logger.warning("no cases found")

    for key in sorted(common_keys):
        yield (key, inputs[key], outputs[key])


def merge_testcase_files(
    inputs: dict[str, pathlib.Path],
    outputs: dict[str, pathlib.Path],
) -> Iterator[TestCaseFile]:
    for name, i, o in enumerate_inouts(inputs, outputs):
        yield TestCaseFile(name=name, input_path=i, output_path=o)


def _casename(path: pathlib.Path, *, directory: pathlib.Path) -> str:
    return path.relative_to(directory).with_suffix("").as_posix()


def iter_testcases(
    *, directory: pathlib.Path, recursive: bool = False
) -> Iterator[TestCaseFile]:
    inputs: dict[str, pathlib.Path] = {}
    outputs: dict[str, pathlib.Path] = {}
    pre = "**/" if recursive else ""

    for path in directory.glob(pre + "*.in"):
        if path.is_file():
            inputs[_casename(path, directory=directory)] = path
    for path in directory.glob(pre + "*.out"):
        if path.is_file():
            outputs[_casename(path, directory=directory)] = path

    return merge_testcase_files(inputs, outputs)


def _name_to_filename(name: str, ext: str):
    return pathlib.Path(name).with_suffix(f".{ext}").name


def save_testcases(samples: Iterable[TestCaseData], *, directory: pathlib.Path):
    for sample in samples:
        for data, ext in [(sample.input_data, "in"), (sample.output_data, "out")]:
            path = directory / _name_to_filename(sample.name, ext)

            if path.exists():
                logger.error("Failed to download since file already exists: %s", path)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_bytes(data)
            logger.debug("saved to: %s", path)



import shutil
from typing import Any

from .integration_data import IntegrationData


class JavaData(IntegrationData):
    def check_envinronment(self) -> bool:
        return bool(shutil.which("javac"))

    def expected_verify_json(self) -> dict[str, Any]:
        return {
            "files": {
                "examples/Aplusb.java": {
                    "additonal_sources": [],
                    "dependencies": [
                        "examples/Aplusb.java",
                        "examples/Aplusb_main.java",
                        "examples/Aplusb_test.java",
                        "examples/HelloWorld.java",
                        "examples/HelloWorld_test.java",
                    ],
                    "document_attributes": {},
                    "verification": [],
                },
                "examples/Aplusb_main.java": {
                    "additonal_sources": [],
                    "dependencies": [
                        "examples/Aplusb.java",
                        "examples/Aplusb_main.java",
                        "examples/Aplusb_test.java",
                        "examples/HelloWorld.java",
                        "examples/HelloWorld_test.java",
                    ],
                    "document_attributes": {"STANDALONE": ""},
                    "verification": [
                        {
                            "command": ["java", "examples.Aplusb_main"],
                            "compile": [
                                "javac",
                                str(self.targets_path / "examples/Aplusb_main.java"),
                            ],
                            "name": "Java",
                            "tempdir": f"{self.config_dir_path / 'cache/standalone/382841ad26b555d39a8784691c59fce8'}",
                            "type": "command",
                        }
                    ],
                },
                "examples/Aplusb_test.java": {
                    "additonal_sources": [],
                    "dependencies": [
                        "examples/Aplusb.java",
                        "examples/Aplusb_main.java",
                        "examples/Aplusb_test.java",
                        "examples/HelloWorld.java",
                        "examples/HelloWorld_test.java",
                    ],
                    "document_attributes": {
                        "PROBLEM": "https://judge.yosupo.jp/problem/aplusb"
                    },
                    "verification": [
                        {
                            "command": ["java", "examples.Aplusb_test"],
                            "compile": [
                                "javac",
                                str(self.targets_path / "examples/Aplusb_test.java"),
                            ],
                            "name": "Java",
                            "problem": "https://judge.yosupo.jp/problem/aplusb",
                            "type": "problem",
                        }
                    ],
                },
                "examples/HelloWorld.java": {
                    "additonal_sources": [],
                    "dependencies": [
                        "examples/Aplusb.java",
                        "examples/Aplusb_main.java",
                        "examples/Aplusb_test.java",
                        "examples/HelloWorld.java",
                        "examples/HelloWorld_test.java",
                    ],
                    "document_attributes": {},
                    "verification": [],
                },
                "examples/HelloWorld_test.java": {
                    "additonal_sources": [],
                    "dependencies": [
                        "examples/Aplusb.java",
                        "examples/Aplusb_main.java",
                        "examples/Aplusb_test.java",
                        "examples/HelloWorld.java",
                        "examples/HelloWorld_test.java",
                    ],
                    "document_attributes": {
                        "PROBLEM": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A"
                    },
                    "verification": [
                        {
                            "command": ["java", "examples.HelloWorld_test"],
                            "compile": [
                                "javac",
                                str(
                                    self.targets_path / "examples/HelloWorld_test.java"
                                ),
                            ],
                            "name": "Java",
                            "problem": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A",
                            "type": "problem",
                        }
                    ],
                },
            },
        }

    def expected_verify_result(self) -> dict[str, Any]:
        return {
            "files": {
                "examples/Aplusb_main.java": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 5854.0,
                            "last_execution_time": "1973-11-01T17:29:53.180000+06:00",
                            "status": "success",
                            "verification_name": "Java",
                        }
                    ],
                },
                "examples/Aplusb_test.java": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 1440.0,
                            "heaviest": 331.0,
                            "last_execution_time": "2013-11-08T23:55:48.570000-05:00",
                            "slowest": 144.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 5.84,
                                    "memory": 18.62,
                                    "name": "example_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 8.73,
                                    "memory": 11.04,
                                    "name": "example_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 8.11,
                                    "memory": 84.96,
                                    "name": "random_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.84,
                                    "memory": 71.78,
                                    "name": "random_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 5.44,
                                    "memory": 56.52,
                                    "name": "random_02",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 5.1,
                                    "memory": 94.73,
                                    "name": "random_03",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.52,
                                    "memory": 85.38,
                                    "name": "random_04",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 2.03,
                                    "memory": 95.06,
                                    "name": "random_05",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 4.13,
                                    "memory": 25.69,
                                    "name": "random_06",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.88,
                                    "memory": 2.09,
                                    "name": "random_07",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.73,
                                    "memory": 79.09,
                                    "name": "random_08",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 1.31,
                                    "memory": 63.99,
                                    "name": "random_09",
                                    "status": "AC",
                                },
                            ],
                            "verification_name": "Java",
                        }
                    ],
                },
                "examples/HelloWorld_test.java": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 8002.0,
                            "heaviest": 302.0,
                            "last_execution_time": "2007-06-15T09:14:19.970000+10:00",
                            "slowest": 800.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 3.73,
                                    "memory": 70.0,
                                    "name": "judge_data",
                                    "status": "AC",
                                }
                            ],
                            "verification_name": "Java",
                        }
                    ],
                },
            },
            "total_seconds": 3781.68,
        }



import pathlib
import tempfile
from collections.abc import Generator

import pytest


@pytest.fixture(scope="session")
def integration_test_data_dir() -> pathlib.Path:
    return pathlib.Path(__file__).parent.parent.parent / "integration_test_data"


@pytest.fixture
def testtemp() -> Generator[pathlib.Path, None, None]:
    with tempfile.TemporaryDirectory() as d:
        yield pathlib.Path(d).resolve()



# Python Version: 3.x
import functools
import pathlib
from collections.abc import Generator
from logging import getLogger

from pydantic import BaseModel, Field

from competitive_verifier.util import read_text_normalized

from .base import Language, LanguageEnvironment, OjVerifyLanguageConfig

logger = getLogger(__name__)

DEFAULT_COMPILE_TO = "cpp"
DEFAULT_NIM_FLAGS = ["-d:release", "--opt:speed"]


class OjVerifyNimConfigEnv(BaseModel):
    compile_to: str | None
    NIMFLAGS: list[str] | None


class OjVerifyNimConfig(OjVerifyLanguageConfig):
    environments: list[OjVerifyNimConfigEnv] = Field(
        default_factory=list[OjVerifyNimConfigEnv]
    )


class NimLanguageEnvironment(LanguageEnvironment):
    compile_to: str
    nim_flags: list[str]

    def __init__(self, *, compile_to: str, nim_flags: list[str]):
        self.compile_to = compile_to
        self.nim_flags = nim_flags

    @property
    def name(self) -> str:
        return "Nim"

    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> list[str]:
        return [
            "nim",
            self.compile_to,
            "-p:.",
            f"-o:{tempdir / 'a.out'!s}",
            f"--nimcache:{tempdir!s}",
            *self.nim_flags,
            str(path),
        ]

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        return str(tempdir / "a.out")


def _parse_path(lines: list[str]) -> Generator[str, None, None]:
    for line in lines:
        l = line.strip()
        if l.startswith("include"):
            yield from l[7:].strip().split(",")
        elif l.startswith("import"):
            l = l[6:]
            i = l.find(" except ")
            if i >= 0:
                l = l[:i]
            yield from l.split(",")
        elif l.startswith("from"):
            i = l.find(" import ")
            if i >= 0:
                yield l[4 : i - 1]


def _unquote_path(p: str) -> pathlib.Path:
    p = p.strip()
    if p.startswith('"'):
        p = p[1 : len(p) - 1]
    else:
        p += ".nim"
    return pathlib.Path(p)


@functools.cache
def _list_direct_dependencies(
    path: pathlib.Path,
    *,
    basedir: pathlib.Path,
) -> list[pathlib.Path]:
    dependencies = [path.resolve()]
    for item in _parse_path(read_text_normalized(basedir / path).splitlines()):
        item_ = _unquote_path(item)
        if item_.exists():
            dependencies.append(item_)
    return list(set(dependencies))


class NimLanguage(Language):
    config: OjVerifyNimConfig = Field(default_factory=OjVerifyNimConfig)

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        dependencies: list[pathlib.Path] = []
        visited: set[pathlib.Path] = set()
        stk = [path.resolve()]
        while stk:
            path = stk.pop()
            if path in visited:
                continue
            visited.add(path)
            for child in _list_direct_dependencies(path, basedir=basedir):
                dependencies.append(child)
                stk.append(child)
        return list(set(dependencies))

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[NimLanguageEnvironment]:
        if self.config.environments:
            return [
                NimLanguageEnvironment(
                    compile_to=DEFAULT_COMPILE_TO,
                    nim_flags=DEFAULT_NIM_FLAGS,
                )
            ]
        return [
            NimLanguageEnvironment(
                compile_to=env.compile_to or DEFAULT_COMPILE_TO,
                nim_flags=env.NIMFLAGS or DEFAULT_NIM_FLAGS,
            )
            for env in self.config.environments
        ]



import math
import pathlib
from argparse import ArgumentParser
from functools import cached_property
from logging import getLogger
from typing import Literal

from pydantic import Field, field_validator

from competitive_verifier import github
from competitive_verifier.arg import (
    IgnoreErrorArguments,
    VerboseArguments,
    VerifyFilesJsonArguments,
    WriteSummaryArguments,
)
from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import VerificationInput, VerifyCommandResult

from .verifier import SplitState, Verifier

logger = getLogger(__name__)


class Verify(
    WriteSummaryArguments,
    IgnoreErrorArguments,
    VerifyFilesJsonArguments,
    VerboseArguments,
):
    subcommand: Literal["verify"] = Field(
        default="verify",
        description="Verify library",
    )
    timeout: float = math.inf
    default_tle: float | None = None
    default_mle: float | None = None

    prev_result: pathlib.Path | None = None

    download: bool = True

    output: pathlib.Path | None = None

    split: int | None = None
    split_index: int | None = None

    def read_prev_result(self) -> VerifyCommandResult | None:
        if not self.prev_result:
            return None
        try:
            return VerifyCommandResult.parse_file_relative(self.prev_result)
        except Exception:
            logger.warning(
                "Failed to parse prev_result: %s",
                self.prev_result,
                extra={"github": GitHubMessageParams(file=self.prev_result)},
            )

    def write_result(self, result: VerifyCommandResult):
        super().write_result(result)

        result_json = result.model_dump_json(exclude_none=True)
        print(result_json)

        if self.output:
            self.output.parent.mkdir(parents=True, exist_ok=True)
            self.output.write_text(result_json, encoding="utf-8")

    @field_validator("timeout", mode="after")
    @classmethod
    def timeout_zero_equals_inf(cls, value: float) -> float:
        if value == 0:
            return math.inf
        return value

    @cached_property
    def split_state(self) -> SplitState | None:
        split = self.split
        split_index = self.split_index
        match (split_index, split):
            case (int(), int()):
                if split <= 0:
                    raise ValueError("--split must be greater than 0.")
                if not (0 <= split_index < split):
                    raise ValueError(
                        "--split-index must be greater than 0 and less than --split."
                    )
                return SplitState(size=split, index=split_index)
            case (None, int()):
                raise ValueError("--split argument requires --split-index argument.")
            case (int(), None):
                raise ValueError("--split-index argument requires --split argument.")
            case _:
                return None

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "--timeout",
            type=float,
            default=math.inf,
            help="Timeout seconds. if value is zero, it is same to math.inf.",
        )
        parser.add_argument(
            "--tle",
            dest="default_tle",
            type=float,
            default=None,
            help="Threshold seconds to be TLE",
        )
        parser.add_argument(
            "--mle",
            dest="default_mle",
            type=float,
            default=None,
            help="Threshold memory usage (MB) to be MLE",
        )
        parser.add_argument(
            "--prev-result",
            type=pathlib.Path,
            required=False,
            help="Previous result json file",
        )

        parser.add_argument(
            "--no-download",
            action="store_false",
            dest="download",
            help="Suppress `oj download`",
        )
        parser.add_argument(
            "--output",
            "-o",
            type=pathlib.Path,
            required=False,
            help="The output file for which verifier saves the result json.",
        )
        parallel_group = parser.add_argument_group("parallel")
        parallel_group.add_argument(
            "--split",
            type=int,
            help="Parallel job size",
            required=False,
        )
        parallel_group.add_argument(
            "--split-index",
            type=int,
            help="Parallel job index",
            required=False,
        )

    def _run(self) -> bool:
        logger.debug("arguments:%s", self)
        logger.info("verify_files_json=%s", self.verify_files_json)
        verifications = VerificationInput.parse_file_relative(self.verify_files_json)
        prev_result = self.read_prev_result()

        verifier = Verifier(
            verifications,
            use_git_timestamp=github.env.is_in_github_actions(),
            timeout=self.timeout,
            default_tle=self.default_tle,
            default_mle=self.default_mle,
            prev_result=prev_result,
            split_state=self.split_state,
        )
        result = verifier.verify(download=self.download)
        self.write_result(result)

        is_success = result.is_success()

        if is_success:
            logger.info("success!")
        else:
            logger.warning("not success!")

        return is_success or self.ignore_error



import os
import pathlib
import platform
import shlex
import shutil
from logging import getLogger
from typing import Any

from pydantic import BaseModel, Field

from competitive_verifier.exec import command_stdout
from competitive_verifier.log import GitHubMessageParams

from . import special_comments
from .base import Language, LanguageEnvironment, OjVerifyLanguageConfig
from .cplusplus_bundle import Bundler

# ruff: noqa: N803

logger = getLogger(__name__)


class OjVerifyCPlusPlusConfigEnv(BaseModel):
    CXX: str
    CXXFLAGS: list[str] | None = None


class OjVerifyCPlusPlusConfig(OjVerifyLanguageConfig):
    environments: list[OjVerifyCPlusPlusConfigEnv] | None = None


class CPlusPlusLanguageEnvironment(LanguageEnvironment):
    cxx: pathlib.Path
    cxx_flags: list[str]

    def __init__(self, *, CXX: pathlib.Path, CXXFLAGS: list[str]):
        self.cxx = CXX
        self.cxx_flags = CXXFLAGS

    @property
    def name(self) -> str:
        return self.cxx.name

    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> list[str]:
        return [
            str(self.cxx),
            *self.cxx_flags,
            "-I",
            str(basedir),
            "-o",
            str(tempdir / "a.out"),
            str(path),
        ]

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        return str(tempdir / "a.out")

    def is_clang(self) -> bool:
        return "clang++" in self.cxx.name

    def is_gcc(self) -> bool:
        return not self.is_clang() and "g++" in self.cxx.name


def _cplusplus_list_depending_files(
    path: pathlib.Path,
    *,
    CXX: pathlib.Path,
    CXXFLAGS: list[str],
) -> list[pathlib.Path]:
    is_windows = platform.uname().system == "Windows"
    command = [str(CXX), *CXXFLAGS, "-MM", str(path)]
    try:
        data = command_stdout(command)
    except Exception:
        logger.exception(
            "failed to analyze dependencies with %s: %s  (hint: Please check #include directives of the file and its dependencies."
            " The paths must exist, must not contain '\\', and must be case-sensitive.)",
            CXX,
            path,
            exc_info=False,
        )
        raise
    logger.debug("dependencies of %s: %r", path, data)
    makefile_rule = shlex.split(
        data.strip().replace("\\\n", "").replace("\\\r\n", ""),
        posix=not is_windows,
    )
    return [pathlib.Path(path).resolve() for path in makefile_rule[1:]]


def _cplusplus_list_defined_macros(
    path: pathlib.Path, *, CXX: pathlib.Path, CXXFLAGS: list[str]
) -> dict[str, str]:
    command = [str(CXX), *CXXFLAGS, "-dM", "-E", str(path)]
    data = command_stdout(command)
    define: dict[str, str] = {}
    for line in data.splitlines():
        assert line.startswith("#define ")
        a, _, b = line[len("#define ") :].partition(" ")
        if (b.startswith('"') and b.endswith('"')) or (
            b.startswith("'") and b.endswith("'")
        ):
            b = b[1:-1]
        define[a] = b
    return define


_NOT_SPECIAL_COMMENTS = "*NOT_SPECIAL_COMMENTS*"
_PROBLEM = "PROBLEM"
_IGNORE = "IGNORE"
_IGNORE_IF_CLANG = "IGNORE_IF_CLANG"
_IGNORE_IF_GCC = "IGNORE_IF_GCC"
_ERROR = "ERROR"
_STANDALONE = "STANDALONE"


class CPlusPlusLanguage(Language):
    config: OjVerifyCPlusPlusConfig = Field(default_factory=OjVerifyCPlusPlusConfig)

    def _list_environments(self) -> list[CPlusPlusLanguageEnvironment]:
        default_CXXFLAGS = ["--std=c++17", "-O2", "-Wall", "-g"]  # noqa: N806
        if platform.system() == "Windows" or "CYGWIN" in platform.system():
            default_CXXFLAGS.append("-Wl,-stack,0x10000000")
        if platform.system() == "Darwin":
            default_CXXFLAGS.append("-Wl,-stack_size,0x10000000")
        if (
            platform.uname().system == "Linux"
            and "Microsoft" in platform.uname().release
        ):
            default_CXXFLAGS.append("-fsplit-stack")

        if "CXXFLAGS" in os.environ and not self.config.environments:
            logger.warning(
                "Usage of $CXXFLAGS envvar to specify options is deprecated and will be removed soon",
                extra={"github": GitHubMessageParams()},
            )
            default_CXXFLAGS = shlex.split(os.environ["CXXFLAGS"])  # noqa: N806

        envs: list[CPlusPlusLanguageEnvironment] = []
        if self.config.environments:
            # configured: use specified CXX & CXXFLAGS
            envs.extend(
                CPlusPlusLanguageEnvironment(
                    CXX=pathlib.Path(env.CXX),
                    CXXFLAGS=env.CXXFLAGS or default_CXXFLAGS,
                )
                for env in self.config.environments
            )

        elif "CXX" in os.environ:
            # old-style: 以前は $CXX を使ってたけど設定ファイルに移行したい
            logger.warning(
                "Usage of $CXX envvar to restrict compilers is deprecated and will be removed soon",
                extra={"github": GitHubMessageParams()},
            )
            envs.append(
                CPlusPlusLanguageEnvironment(
                    CXX=pathlib.Path(os.environ["CXX"]), CXXFLAGS=default_CXXFLAGS
                )
            )

        else:
            # default: use found compilers
            for name in ("g++", "clang++"):
                path = shutil.which(name)
                if path is not None:
                    envs.append(
                        CPlusPlusLanguageEnvironment(
                            CXX=pathlib.Path(path), CXXFLAGS=default_CXXFLAGS
                        )
                    )

        if not envs:
            raise RuntimeError("No C++ compilers found")
        return envs

    def list_attributes(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> dict[str, Any]:
        attributes: dict[str, Any] = {}

        comments = special_comments.list_special_comments(path.resolve())
        if comments:
            attributes.update(comments)
        else:
            # use old-style if special comments not found
            # #define PROBLEM "https://..." の形式は複数 environments との相性がよくない。あと遅い
            attributes[_NOT_SPECIAL_COMMENTS] = ""
            all_ignored = True
            for env in self._list_environments():
                macros = _cplusplus_list_defined_macros(
                    path.resolve(),
                    CXX=env.cxx,
                    CXXFLAGS=[*env.cxx_flags, "-I", str(basedir)],
                )

                # convert macros to attributes
                if _IGNORE not in macros:
                    if _STANDALONE in macros:
                        attributes[_STANDALONE] = ""

                    for key in [_PROBLEM, _ERROR]:
                        if all_ignored:
                            # the first non-ignored environment
                            if key in macros:
                                attributes[key] = macros[key]
                        else:
                            assert attributes.get(key) == macros.get(key)
                    all_ignored = False
                elif env.is_gcc():
                    attributes[_IGNORE_IF_GCC] = ""
                elif env.is_clang():
                    attributes[_IGNORE_IF_CLANG] = ""
                else:
                    attributes[_IGNORE] = ""
            if all_ignored:
                attributes[_IGNORE] = ""

        attributes.setdefault("links", [])
        attributes["links"].extend(special_comments.list_embedded_urls(path))
        return attributes

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        env = self._list_environments()[0]
        return _cplusplus_list_depending_files(
            path.resolve(),
            CXX=env.cxx,
            CXXFLAGS=[*env.cxx_flags, "-I", str(basedir)],
        )

    def bundle(self, path: pathlib.Path, *, basedir: pathlib.Path) -> bytes | None:
        include_paths: list[pathlib.Path] = [basedir]
        assert isinstance(include_paths, list)
        bundler = Bundler(iquotes=include_paths)
        bundler.update(path)
        return bundler.get()

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[CPlusPlusLanguageEnvironment]:
        attributes = self.list_attributes(path, basedir=basedir)
        envs: list[CPlusPlusLanguageEnvironment] = []
        for env in self._list_environments():
            if env.is_gcc() and _IGNORE_IF_GCC in attributes:
                continue
            if env.is_clang() and _IGNORE_IF_CLANG in attributes:
                continue
            envs.append(env)
        return envs



import os
import pathlib

from pytest_mock import MockerFixture

from competitive_verifier import config


def test_config_dir(mocker: MockerFixture):
    mocker.patch.dict(os.environ, {}, clear=True)

    assert config.get_config_dir() == pathlib.Path(".competitive-verifier")
    assert config.get_cache_dir() == pathlib.Path(".competitive-verifier/cache")
    assert config.get_problem_cache_dir() == pathlib.Path(
        ".competitive-verifier/cache/problems"
    )

    mocker.patch.dict(
        os.environ, {"COMPETITIVE_VERIFY_CONFIG_PATH": "/root/foo"}, clear=True
    )

    assert config.get_config_dir() == pathlib.Path("/root/foo")
    assert config.get_cache_dir() == pathlib.Path("/root/foo/cache")
    assert config.get_problem_cache_dir() == pathlib.Path("/root/foo/cache/problems")



import logging
import pathlib
from datetime import datetime, timedelta, timezone

import pytest
from pydantic import ValidationError

from competitive_verifier.app import Check
from competitive_verifier.models import (
    FileResult,
    ResultStatus,
    VerificationResult,
    VerifyCommandResult,
)
from tests import LogComparer

test_check_params: list[tuple[dict[str, VerifyCommandResult], str, bool]] = [
    (
        {
            "file1.json": VerifyCommandResult(
                total_seconds=1.25,
                files={
                    pathlib.Path("file1.txt"): FileResult(
                        newest=False,
                        verifications=[
                            VerificationResult(
                                verification_name="v1",
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime(
                                    2025,
                                    9,
                                    21,
                                    8,
                                    9,
                                    10,
                                    1131,
                                    tzinfo=timezone(timedelta(hours=+9)),
                                ),
                                elapsed=1.2,
                            ),
                        ],
                    ),
                    pathlib.Path("file2.txt"): FileResult(
                        newest=True,
                        verifications=[
                            VerificationResult(
                                verification_name="v2",
                                status=ResultStatus.SKIPPED,
                                last_execution_time=datetime(
                                    2025,
                                    9,
                                    21,
                                    8,
                                    9,
                                    11,
                                    4253,
                                    tzinfo=timezone(timedelta(hours=-9)),
                                ),
                                elapsed=1.3,
                            ),
                        ],
                    ),
                },
            ),
        },
        "success: 1\nfailure: 0\nskipped: 1\n",
        True,
    ),
    (
        {
            "file1.json": VerifyCommandResult(
                total_seconds=1.25,
                files={
                    pathlib.Path("file1.txt"): FileResult(
                        newest=False,
                        verifications=[
                            VerificationResult(
                                verification_name="v1",
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime(
                                    2025,
                                    9,
                                    21,
                                    8,
                                    9,
                                    10,
                                    1131,
                                    tzinfo=timezone(timedelta(hours=+9)),
                                ),
                                elapsed=1.2,
                            ),
                        ],
                    ),
                    pathlib.Path("file2.txt"): FileResult(
                        newest=True,
                        verifications=[
                            VerificationResult(
                                verification_name="v2",
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime(
                                    2025,
                                    9,
                                    21,
                                    8,
                                    9,
                                    11,
                                    4253,
                                    tzinfo=timezone(timedelta(hours=-9)),
                                ),
                                elapsed=1.3,
                            ),
                        ],
                    ),
                },
            ),
            "file2.json": VerifyCommandResult(
                total_seconds=1.44,
                files={
                    pathlib.Path("file1.txt"): FileResult(
                        newest=True,
                        verifications=[
                            VerificationResult(
                                verification_name="v3",
                                status=ResultStatus.FAILURE,
                                last_execution_time=datetime(
                                    2025,
                                    9,
                                    21,
                                    8,
                                    9,
                                    11,
                                    63474,
                                    tzinfo=timezone.utc,
                                ),
                                elapsed=1.7,
                            ),
                        ],
                    ),
                    pathlib.Path("file2.txt"): FileResult(
                        newest=False,
                        verifications=[
                            VerificationResult(
                                verification_name="v4",
                                status=ResultStatus.FAILURE,
                                last_execution_time=datetime(
                                    2025,
                                    9,
                                    21,
                                    8,
                                    9,
                                    11,
                                    2144,
                                    tzinfo=timezone(timedelta(hours=-1)),
                                ),
                                elapsed=1.9,
                            ),
                        ],
                    ),
                },
            ),
        },
        "success: 1\nfailure: 1\nskipped: 0\n",
        False,
    ),
]


@pytest.mark.parametrize(("files", "expected_out", "expected"), test_check_params)
def test_check(
    files: dict[str, VerifyCommandResult],
    expected_out: str,
    expected: bool,
    testtemp: pathlib.Path,
    capsys: pytest.CaptureFixture[str],
):
    file_list: list[pathlib.Path] = []
    for k, v in files.items():
        p = testtemp / k
        file_list.append(p)
        j = v.model_dump_json()
        p.write_text(j.replace(r"{base}", testtemp.as_posix()))

    assert Check(result_json=file_list).run() == expected
    out, err = capsys.readouterr()
    assert err == ""
    assert out == expected_out


def test_check_error(testtemp: pathlib.Path):
    (testtemp / "ok.json").write_text(
        r"""{
    "files":{}
}"""
    )
    (testtemp / "ng.json").write_text(
        r"""{
    "files":[]
}"""
    )
    with pytest.raises(
        ValidationError, match=r"validation error for VerifyCommandResult"
    ):
        Check(result_json=[testtemp / "ok.json", testtemp / "ng.json"]).run()


def test_log(testtemp: pathlib.Path, caplog: pytest.LogCaptureFixture):
    caplog.at_level(logging.NOTSET)
    (testtemp / "success.json").write_text(
        r"""
{
    "total_seconds": 1.25,
    "files": {
        "file1.txt": {
            "newest": false,
            "verifications": [
                {
                    "elapsed": 1.2,
                    "last_execution_time": "2025-09-21T08:09:10.001131+09:00",
                    "status": "success",
                    "verification_name": "v1"
                }
            ]
        },
        "file2.txt": {
            "newest": true,
            "verifications": [
                {
                    "elapsed": 1.3,
                    "last_execution_time": "2025-09-21T08:09:11.004253-09:00",
                    "status": "success",
                    "verification_name": "v2"
                }
            ]
        }
    }
}"""
    )
    (testtemp / "failure.json").write_text(
        r"""
{
    "total_seconds": 1.25,
    "files": {
        "file1.txt": {
            "newest": false,
            "verifications": [
                {
                    "elapsed": 1.2,
                    "last_execution_time": "2025-09-21T08:09:10.001131+09:00",
                    "status": "success",
                    "verification_name": "v3"
                }
            ]
        },
        "file2.txt": {
            "newest": true,
            "verifications": [
                {
                    "elapsed": 1.3,
                    "last_execution_time": "2025-09-21T08:09:11.004253-09:00",
                    "status": "failure",
                    "verification_name": "v4"
                }
            ]
        }
    }
}"""
    )
    assert Check(verbose=True, result_json=[testtemp / "success.json"]).run()
    assert not caplog.records

    assert not Check(
        verbose=True,
        result_json=[testtemp / "success.json", testtemp / "failure.json"],
    ).run()
    assert caplog.records == [LogComparer("Failure test count: 1", logging.ERROR)]
